# coding: utf-8

# $Id: $


import time
import json
import sys

import dateutil.parser
from django.core.signals import request_started, request_finished
from django.db import connections
import redis
from amqp import Connection
from alco.collector.defaults import ALCO_SETTINGS

from alco.collector import keys


class Collector(object):

    def __init__(self, index):
        self.index = index
        self.cancelled = False
        self.transport = self.protocol = None
        self.messages = []
        self.block_size = 1000
        self.exchange = "logstash"
        self.current_date = None

    def cancel(self):
        self.cancelled = True

    def connect(self):
        self.amqp = Connection(**ALCO_SETTINGS['RABBITMQ'])
        self.redis = redis.Redis(**ALCO_SETTINGS['REDIS'])
        self.conn = connections[ALCO_SETTINGS['SPHINX_DATABASE_NAME']]

    def __call__(self, *args, **kwargs):
        try:
            self.connect()
            self.declare_queue()
            channel = self.amqp.channel()
            channel.basic_consume("logstash", callback=self.process_message, no_ack=True)
            start = time.time()
            while not self.cancelled:
                channel.wait()
                if time.time() - start > 1:
                    self.push_messages()
                    start = time.time()
        except KeyboardInterrupt:
            self.amqp.close()
            sys.exit(0)

    def process_message(self, msg):
        js = msg.body.decode("utf-8")
        data = json.loads(js)
        ts = data.pop('@timestamp')
        data.pop("@version")
        msg = data.pop('message')
        seq = data.pop('seq', 0)
        dt = dateutil.parser.parse(ts)
        result = {
            'ts': time.mktime(dt.timetuple()),
            'ms': dt.microsecond,
            'seq': seq,
            'message': msg,
            'js': json.dumps(data),
            'data': data
        }
        self.messages.append(result)
        d = dt.date()
        if not self.current_date:
            self.current_date = d
        if d != self.current_date:
            self.current_date = d
            self.push_messages()
        if len(self.messages) >= self.block_size:
            self.push_messages()

    def declare_queue(self):
        channel = self.amqp.channel()
        channel.exchange_declare(exchange=self.exchange, type='fanout',
                                 durable=True, auto_delete=False)
        channel.queue_declare(self.index.queue_name, durable=True,
                              auto_delete=False)
        channel.queue_bind(self.index.queue_name,
                           exchange=self.exchange,
                           routing_key=self.index.routing_key)

    def push_messages(self):
        try:
            request_started.send(None, environ=None)
            self._push_messages()
        finally:
            request_finished.send(None)

    def _push_messages(self):
        messages, self.messages = self.messages, []
        if not messages:
            return
        key = keys.KEY_SEQUENCE.format(index=self.index.name)
        max_pk = self.redis.incrby(key, len(messages))
        min_pk = max_pk - len(messages)

        columns = dict()

        suffix = self.current_date.strftime("%Y%m%d")
        name = "%s_%s" % (self.index.name, suffix)
        query = "REPLACE INTO %s (id, ts, ms, seq, js, logline) VALUES " % name
        rows = []
        args = []
        for pk, data in zip(range(min_pk, max_pk), messages):
            # saving seen columns to LoggerColumn model, collecting unique
            # values for caching in redis
            for key, value in data['data'].items():
                columns.setdefault(key, set())
                if not isinstance(value, (bool, int, float, str)):
                    continue
                columns[key].add(value)

            rows.append("(%s, %s, %s, %s, %s, %s)")
            args.extend((pk, data['ts'], data['ms'], data['seq'], data['js'],
                         data['message']))
        query += ','.join(rows)

        existing = self.index.loggercolumn_set.all()
        filtered = filter(lambda c: c.filtered, existing)

        existing = [c.name for c in existing]
        filtered = [c.name for c in filtered]
        new_values = set(columns.keys() - set(existing))

        for column in filtered:
            values = columns.get(column)
            if not values:
                continue
            key = keys.KEY_COLUMN_VALUES.format(index=self.index.name,
                                                column=column)
            self.redis.sadd(key, *values)

        for column in new_values:
            self.index.loggercolumn_set.create(name=column)

        for _ in range(3):
            try:
                c = self.conn.cursor()
                c.execute(query, args)
                print(c.rowcount)
                c.close()
            except Exception as e:
                pass
            else:
                break







# coding: utf-8

# $Id: $
import os
from time import sleep
from daemon import DaemonContext
from daemon.pidfile import TimeoutPIDLockFile

from django.core.management import BaseCommand
from django.utils.log import getLogger
from multiprocessing import Process
import signal
from alco.collector.collector import Collector
from alco.collector.models import LoggerIndex

logger = getLogger(__name__)


class Command(BaseCommand):
    def __init__(self, stdout=None, stderr=None, no_color=False):
        super().__init__(stdout=None, stderr=None, no_color=False)
        self.processes = {}
        self.started = False

    def add_arguments(self, parser):
        parser.add_argument(
            '--no-daemon', action='store_false', dest='daemon',
            default=True, help="Don't daemonize process")
        parser.add_argument(
            '--pidfile', action='store', dest='pidfile',
            default="collector.pid", help="pidfile location")

    def handle(self, *args, **options):
        if not options['daemon']:
            return self.start_worker_pool()

        path = os.path.join(os.getcwd(), options['pidfile'])
        pidfile = TimeoutPIDLockFile(path)

        with DaemonContext(pidfile=pidfile):
            logger.info("daemonized")
            self.start_worker_pool()

    def start_worker_pool(self):
        indices = list(LoggerIndex.objects.all())
        n = 0
        for index in indices:
            c = Collector(index)
            p = Process(target=c)
            p.start()
            self.processes[n] = (p, index)
            n += 1
        self.started = True
        signal.signal(signal.SIGINT, self.handle_sigint)

        while len(self.processes) > 0:
            if not self.started:
                for n in self.processes:
                    (p, index) = self.processes[n]
                    os.kill(p.pid, signal.SIGTERM)
            sleep(0.5)
            for n in list(self.processes):
                (p, index) = self.processes[n]
                if p.exitcode is None:
                    if not p.is_alive() and self.started:
                        # Not finished and not running
                        os.kill(p.pid, signal.SIGKILL)
                        c = Collector(index)
                        p = Process(target=c)
                        p.start()
                        self.processes[n] = (p, index)
                elif p.exitcode < 0 and self.started:
                    print('Process Ended with an error or a terminate', index)
                    c = Collector(index)
                    p = Process(target=c)
                    p.start()
                    self.processes[n] = (p, index)
                else:
                    print(index, 'finished')
                    p.join()
                    del self.processes[n]

    def handle_sigint(self, sig_num, frame):
        self.started = False




# coding: utf-8

# $Id: $
from django.utils import six
from rest_framework.filters import BaseFilterBackend
from rest_framework.settings import api_settings

from sphinxsearch.utils import sphinx_escape


class SphinxSearchFilter(BaseFilterBackend):
    # The URL query parameter used for the search.
    search_param = api_settings.SEARCH_PARAM

    def get_search_terms(self, request, escape=True):
        """
        Search terms are set by a ?search=... query parameter,
        and may be comma and/or whitespace delimited.
        """
        params = request.query_params.get(self.search_param, '')
        if not params or params == " ":
            return []
        if escape:
            return list(map(sphinx_escape, params.replace(',', ' ').split()))
        return params.replace(',', ' ').split()

    def construct_search(self, field_name):
        return '@%s ("%%s")' % field_name

    def filter_queryset(self, request, queryset, view):
        search_fields = getattr(view, 'search_fields', None)

        if not search_fields:
            return queryset

        search_terms = self.get_search_terms(request)
        if not search_terms:
            return queryset

        orm_lookups = [self.construct_search(six.text_type(search_field))
                       for search_field in search_fields]

        search_query = ' & '.join(search_terms)
        match_expression = '|'.join(lookup % search_query
                                    for lookup in orm_lookups)
        queryset = queryset.match(match_expression)
        select = dict(logline_snippet="SNIPPET(logline, %s, 'limit=1000000')")
        params = [' '.join(self.get_search_terms(request, escape=False))]
        return queryset.extra(select=select, select_params=params)
        # FIXME: return qs.options(before_match='<ins>', after_match='</ins>')


class JSONFieldFilter(BaseFilterBackend):

    def get_json_fields(self, request, view):
        json_fields = getattr(view, 'get_json_fields', None)
        if callable(json_fields):
            json_fields = json_fields(request)
        else:
            json_fields = getattr(view, 'json_fields', None)
        return json_fields

    def filter_queryset(self, request, queryset, view):
        json_fields = self.get_json_fields(request, view)
        if not json_fields:
            return queryset
        lookups = self.get_lookups(json_fields, request)
        all_values = set()
        for key, value in lookups.items():
            if not isinstance(value, (list, tuple)):
                value = [value]
            all_values.update(value)
        #match = ' | '.join('"%s"' % sphinx_escape(value) for value in all_values)
        queryset = queryset.filter(**lookups)
        return queryset

    def get_lookups(self, json_fields, request):
        lookups = {}
        for key in json_fields:
            value = request.query_params.get(key +'__in')
            if value is None:
                continue
            lookups[key +'__in'] = value.split(',')

        return lookups



# coding: utf-8

# $Id: $
from rest_framework.generics import ListAPIView

from alco.collector.api.serializers import LoggerIndexSerializer
from alco.collector.models import LoggerIndex


class LoggerIndexView(ListAPIView):
    queryset = LoggerIndex.objects.all()
    serializer_class = LoggerIndexSerializer



# coding: utf-8

# $Id: $
from collections import OrderedDict

from django.core.paginator import Paginator, PageNotAnInteger, EmptyPage, Page, \
    InvalidPage
from django.utils import six
from rest_framework.exceptions import NotFound
from rest_framework.pagination import PageNumberPagination
from rest_framework.response import Response
from alco.collector.defaults import ALCO_SETTINGS


class FeedPage(Page):
    """ Endless feed page.

     Next page is determined by expression len(object_list) > per_page.
     `object_list` contains per_page + 1 object to detect whether next page is
     present.
     """

    @property
    def object_list(self):
        """ Returns page object list without after-end indication element."""
        try:
            return self.__object_list[:self.paginator.per_page]
        except TypeError:
            return list()

    @object_list.setter
    def object_list(self, value):
        self.__object_list = value

    def __init__(self, object_list, number, paginator):
        self.__object_list = None
        super(FeedPage, self).__init__(object_list, number, paginator)

    def __repr__(self):
        return '<Page %s of Unknown>' % (self.number,)

    def has_next(self):
        try:
            return len(self.__object_list) > self.paginator.per_page
        except TypeError:
            return False

    def end_index(self):
        raise NotImplementedError("feed doesn't support count")


class FeedPaginator(Paginator):
    """ Endless feed Django paginator."""

    def _get_num_pages(self):
        """ Needs count."""
        raise NotImplementedError()

    def _get_page_range(self):
        """ Needs count."""
        raise NotImplementedError()

    def _get_count(self):
        """ For endless feed count is not needed."""
        raise NotImplementedError()

    def validate_number(self, number):
        """ Validates the given 1-based page number. """
        try:
            number = int(number)
        except ValueError:
            raise PageNotAnInteger('That page number is not an integer')
        if number < 1:
            raise EmptyPage('That page number is less than 1')
        return number

    count = property(_get_count)

    def page(self, number):
        """ Returns a Page object for the given 1-based page number. """
        number = self.validate_number(number)
        bottom = (number - 1) * self.per_page
        top = bottom + self.per_page
        # top + 1 object is needed for FeedPage to detect next page presence.
        if self.object_list is not None:
            return FeedPage(self.object_list[bottom: top + 1], number, self)
        return FeedPage(None, number, self)


class LogPaginator(PageNumberPagination):
    """ DjangoRestFramework endless paginator."""

    page_size = ALCO_SETTINGS['LOG_PAGE_SIZE']

    def paginate_queryset(self, queryset, request, view=None):
        """
        Paginate a queryset if required, either returning a
        page object, or `None` if pagination is not configured for this view.
        """
        self._handle_backwards_compat(view)

        page_size = self.get_page_size(request)
        if not page_size:
            return None

        paginator = FeedPaginator(queryset, page_size)
        page_number = request.query_params.get(self.page_query_param, 1)
        if page_number in self.last_page_strings:
            page_number = paginator.num_pages

        try:
            self.page = paginator.page(page_number)
        except InvalidPage as exc:
            msg = self.invalid_page_message.format(
                page_number=page_number, message=six.text_type(exc)
            )
            raise NotFound(msg)

        if self.page.has_next() > 1 and self.template is not None:
            # The browsable API should display pagination controls.
            self.display_page_controls = True

        self.request = request
        return list(self.page)

    def get_paginated_response(self, data):
        return Response(OrderedDict([
            ('next', self.get_next_link()),
            ('previous', self.get_previous_link()),
            ('results', data)
        ]))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('collector', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='loggercolumn',
            name='display',
            field=models.BooleanField(default=True),
        ),
    ]



# coding: utf-8

# $Id: $
from pprint import pformat

from django import template
import math
from django.template import Node
import re
from django.utils.encoding import force_text
from django.utils.functional import allow_lazy
import six

register = template.Library()


@register.inclusion_tag('grep/tags/filter_item.html')
def filter_item(heading_id, collapse_id, field, title, items):
    return locals()


@register.inclusion_tag('grep/tags/columns_item.html')
def columns_item(heading_id, collapse_id, field, title, items):
    return locals()


@register.inclusion_tag('grep/tags/dates_item.html')
def dates_item(heading_id, collapse_id, field, title, items):
    return locals()


def strip_lines_between_tags(value):
    """Returns the given HTML with spaces between tags removed."""
    return re.sub(r'>\s+<([^/])', '><\\1', force_text(value))
strip_lines_between_tags = allow_lazy(strip_lines_between_tags, six.text_type)


class LinelessNode(Node):
    def __init__(self, nodelist):
        self.nodelist = nodelist

    def render(self, context):
        return strip_lines_between_tags(self.nodelist.render(context).strip())



@register.tag
def lineless(parser, token):
    nodelist = parser.parse(('endlineless',))
    parser.delete_first_token()
    return LinelessNode(nodelist)



# coding: utf-8

# $Id: $


from django.contrib import admin
from alco.grep.models import Shortcut


class ShortcutAdmin(admin.ModelAdmin):
    list_display = ('name', 'description')


admin.site.register(Shortcut, ShortcutAdmin)




# coding: utf-8

# $Id: $

# coding: utf-8

# $Id: server.py 2553 2014-06-19 12:37:13Z stikhonov $
import os
import signal
import asyncio

from dvasya.server import ChildProcess, Supervisor, Worker

from alco.collector.collector import Collector


class CollectorChildProcess(ChildProcess):

    def before_loop(self):
        super().before_loop()
        self.loop.add_signal_handler(signal.SIGHUP, self.reload_config)
        self.logger.debug("Test logging")

    def start(self):
        """ Запускает полезную нагрузку worker-процесса."""
        # В отличие от базового класса, не запускает socket-server.
        # Вместо этого запускает цикл сбора данных.
        self.loop = loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        def stop():
            self.loop.stop()
            os._exit(0)

        loop.add_signal_handler(signal.SIGINT, stop)

        self.before_loop()

        self.start_collector()

        asyncio.get_event_loop().run_forever()
        os._exit(0)

    def start_collector(self):
        """ Инициирует циклы опроса серверов. """
        Collector.launch(self.loop)

    def reload_config(self, *args, **kwargs):
        pass


class CollectorWorker(Worker):
    child_process_class = CollectorChildProcess


class CollectorSupervisor(Supervisor):
    """ Супервизор для процессов сбора логов

    В отличие от dvasya.server.Supervisor не слушает никаких портов.
    """
    worker_class = CollectorWorker

    def open_socket(self):
        """ Убирает базовый метод открытия сокета."""
        return None

    def prefork(self):
        super().prefork()
        self.init_subscribers()

    def init_subscribers(self):
        self.loop.add_signal_handler(signal.SIGHUP, self.reload_config)

    def reload_config(self):
        self.logger.debug("Forced reloading config")
        for worker in self.workers:
            self.logger.debug("KILL %s %s" % (worker.pid, signal.SIGHUP))
            os.kill(worker.pid, signal.SIGHUP)





# coding: utf-8

# $Id: $
from django.utils.functional import cached_property
import django_filters
from rest_framework import filters as rf_filters
from rest_framework.fields import DateTimeField
from rest_framework.generics import ListAPIView

from alco.collector.models import LoggerIndex
from alco.grep.api import filters
from alco.grep.api.pagination import LogPaginator
from alco.grep.models import create_index_model
from alco.grep.api.serializers import LogBaseSerializer


class GrepView(ListAPIView):
    filter_backends = (rf_filters.DjangoFilterBackend,
                       filters.SphinxSearchFilter,
                       filters.JSONFieldFilter)
    search_fields = ('logline',)

    pagination_class = LogPaginator

    @property
    def filter_class(self):
        class TimestampFilter(django_filters.FilterSet):
            start_ts = django_filters.DateTimeFilter(name="ts", lookup_type='gte')
            end_ts = django_filters.NumberFilter(name="ts", lookup_type='lte')

            class Meta:
                model = self.log_model
                fields = ['ts'] + self.index.field_names

        return TimestampFilter

    @cached_property
    def index(self):
        return LoggerIndex.objects.get(name=self.kwargs['logger'])

    @cached_property
    def log_model(self):
        start_ts = self.request.GET.get('start_ts') or ''
        try:
            f = DateTimeField()
            distr = f.to_internal_value(start_ts).strftime('%Y%m%d')
        except Exception:
            distr = None

        return create_index_model(self.index, distr=distr)

    def get_serializer_class(self):
        class Meta:
            model = self.log_model

        serializer_class = type('LogSerializer', (LogBaseSerializer,),
                                {'Meta': Meta})
        return serializer_class

    def get_queryset(self):
        return self.log_model.objects.order_by('ts', 'ms', 'seq')

    def get_json_fields(self, request):
        fields = self.index.filtered_fields
        result = []
        for key, value in request.GET.items():
            field = key.split('__')[0]
            if field in fields:
                result.append(field)
        return result







# coding: utf-8

# $Id: $


from django.views.generic import ListView

from alco.collector.models import LoggerIndex


class LoggerIndexView(ListView):
    model = LoggerIndex
    content_type = 'text/plain'
    template_name = 'collector/sphinx.conf'



# coding: utf-8

# $Id: $

from django.conf import settings

__all__ = ['ALCO_SETTINGS']

# default settings for alco package
ALCO_SETTINGS = {
    # log messaging server
    'RABBITMQ': {
        'host': '127.0.0.1',
        'userid': 'guest',
        'password': 'guest',
        'virtual_host': '/'
    },

    # redis for generating primary key sequences
    'REDIS': {
        'host': '127.0.0.1',
        'db': 0
    },
    # url for fetching sphinx.conf dynamically
    'SPHINX_CONF_URL': 'http://127.0.0.1:8000/',
    # name of django.db.connection for SphinxSearch
    'SPHINX_DATABASE_NAME': 'sphinx',
    # number of results in log view API
    'LOG_PAGE_SIZE': 100
}

# Allow to override some settings from project django conf.

_custom = getattr(settings, 'ALCO_SETTINGS', {})
for k, v in _custom.items():
    ALCO_SETTINGS[k] = v



# coding: utf-8

# $Id: $
import time
from datetime import datetime
from django.db import models
from jsonfield import JSONField

from sphinxsearch.models import SphinxModel, SphinxDateTimeField
from alco.collector.models import LoggerIndex, LoggerColumn


class Shortcut(models.Model):
    name = models.CharField(max_length=30, primary_key=True)
    description = models.CharField(max_length=255, blank=True, default='')
    url = models.CharField(max_length=255)
    index = models.ForeignKey(LoggerIndex)
    default_field = models.ForeignKey(LoggerColumn, blank=True, null=True)

    def __str__(self):
        return self.name


class LogBase(SphinxModel):
    class Meta:
        abstract = True
        ordering = ('ts', 'seq')
        db_table = 'logger'

    ts = SphinxDateTimeField()
    ms = models.IntegerField()
    seq = models.BigIntegerField()
    logline = models.TextField()
    js = JSONField()

    @property
    def datetime(self):
        return self.ts.replace(microsecond=self.ms)

    @property
    def timestamp(self):
        return time.mktime(self.datetime.timetuple())

def create_char_field(c):
    return models.CharField(max_length=255, db_column='js.%s' % c)


def create_index_model(index, distr=None):
    model_name = "%sLog" % index.name.title()
    if distr:
        table = '%s_%s_distr' % (index.name, distr)
    else:
        table = index.name

    class Meta:
        db_table = table

    columns = index.loggercolumn_set.values_list('name', flat=True)

    fields = {c: create_char_field(c) for c in columns}
    fields.update({'Meta': Meta, '__module__': __name__})
    Log = type(model_name, (LogBase,), fields)
    return Log





# coding: utf-8

# $Id: $


from django.conf.urls import url

from alco.grep.views import GrepView, ShortcutView

urlpatterns = [
    url(r'^(?P<name>[\w]+)/$', GrepView.as_view(), name='grep_view'),
    url(r'^shortcut/(?P<name>[\w]+)/(?P<default_value>.+)/$',
        ShortcutView.as_view(), name="shortcut_view"),
    url(r'^shortcut/(?P<name>[\w]+)/$',
        ShortcutView.as_view(), name="shortcut_view"),
]



#!/usr/bin/env python
# coding: utf-8

import sys
import os
import json

settings_module = os.environ.get('DJANGO_SETTINGS_MODULE', 'settings')
settings_module = settings_module.replace('.', '/')
settings_file = settings_module + '.py'
settings_path = os.path.join(os.getcwd(), settings_file)

if sys.version_info[0] > 2:
    from urllib.request import urlopen
    with open(settings_path, 'rb') as f:
        exec(compile(f.read(), settings_file, "exec"))
else:
    # noinspection PyUnresolvedReferences
    from urllib2 import urlopen
    # noinspection PyUnresolvedReferences
    execfile(settings_path)

# noinspection PyUnresolvedReferences
ALCO_HOST = ALCO_SETTINGS['SPHINX_CONF_URL']

r = urlopen(os.path.join(ALCO_HOST, 'collector/sphinx.conf'))
config = r.read().decode('utf-8')

r = urlopen(os.path.join(ALCO_HOST, 'api/collector/indices/?format=json'))
indices = json.loads(r.read().decode('utf-8'))
for idx in indices:
    name = idx['name']
    for dt in idx['index_names']:
        path = os.path.join("/data/sphinx/", name, dt)
        if not os.path.exists(path):
            try:
                os.makedirs(path)
            except OSError:
                pass

print(config)

